
//*******************************************************************
//            RSVP over IPv6 Winsock2 Service Provider
//*******************************************************************
//
// File: rsvpipv6_alloc_and_free_test.cpp
//
// Description: 
//
// Implementation of the functions declared in 
// rsvpipv6_alloc_and_free_test.h.
//
// Copyright 1999, DIT-UPM
//
//*******************************************************************


#define RSVP_IPV6
#include <ws2spi.h>
#include "ws2rsvp.h"

#include "rsvpipv6_alloc_and_free_test.h"


/////////////////////////////////////////////////////////////////////
// Functions implementation
/////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////
// allocates memory for a list of RSVP_FLOWDESC structures.
// initializes whatever necessary.
// Returns pointer to allocated structure or NULL if error
LPRSVP_FLOWDESC
new_flow_desc_list( ULONG numFds ) {

	LPRSVP_FLOWDESC lprf;
	ULONG i; // for looping

	if( (lprf = (LPRSVP_FLOWDESC)malloc( numFds * (sizeof(RSVP_FLOWDESC)) )) 
		== NULL ) {
		// not enough memory:
		return( NULL );
	}

	// Initialize the flowdescriptors' fields:
	for( i=0; i < numFds; i++ ) {
		lprf[i].lpPolicy = NULL;
		lprf[i].numFilts = 0;
		lprf[i].filtList = NULL;
	}

	// Here? -> all went fine:
	return( lprf );
}


/////////////////////////////////////////////////////////////////////
// allocates memory for a list of RSVP_FILTERSPEC structures.
// Returns pointer to allocated structure or NULL if error
LPRSVP_FILTERSPEC
new_RSVP_FILTERSPEC_list( ULONG numfilts ) {

	LPRSVP_FILTERSPEC lprf;

	if( ( lprf = (LPRSVP_FILTERSPEC)
		malloc(numfilts * (sizeof(RSVP_FILTERSPEC))) ) == NULL ) {
		// not enough memory:
		return(NULL);
	}

	//Here->ok:
	return(lprf);

}


/////////////////////////////////////////////////////////////////////
// frees memory allocated for a flow descriptor list of 
// length numFds
int
free_flow_desc_list( ULONG numFds, LPRSVP_FLOWDESC fdList ) {

	ULONG i;
	
	for( i=0; i < numFds; i++ ) {
		if( (fdList[i]).lpPolicy != NULL ) {
			free( (fdList[i]).lpPolicy );
		}
		if( ((fdList[i]).numFilts != 0) && ((fdList[i]).filtList != NULL) ) {
			free( (fdList[i]).filtList );
		}
	}

	free( fdList );

	return( 0 );

}
