// Author: DIT-UPM
//
// Test program for the RSVP over IPv6 provider.
// This provider must be installed (the program does not install it)

#define RSVP_IPV6

#include "Ws2rsvp.h"
#include "rsvpipv6.h" // All applications should include it

#include "rsvpipv6_alloc_and_free_test.h"

#include <ws2ip6.h>
#include <stdio.h>  // fprintf, etc.
 
 
// Number of chained calls to WSAWaitForMultipleEvents:
#define NUM_WAIT 10

// Timeout in milliseconds for waiting events
#define TIMEOUT 2000

#define STR_ADDR_LEN (sizeof("XXXX:XXXX:XXXX:XXXX:XXXX:XXXX:XXXX:XXXX") + 1)
 
///////////////////////////////////////////////////////////////
// Global vars:
char *session_addr, *source_addr;
ushort session_port, source_port;
// token bucket parameters:
FLOAT tb_r, tb_b, tb_p;
LONG tb_m, tb_M;

///////////////////////////////////////////////////////////////
// Usage: print help information
void Usage(char *prog_name) {

	fprintf(stdout, "\nUsage: %s [options]\n", prog_name);
	fprintf(stdout, "\nOptions:\n");
	fprintf(stdout, " -m s|r|b   mode: sender, receiver, both\n");
	fprintf(stdout, " -sa session_address\n");
	fprintf(stdout, " -sp session_port\n");
	fprintf(stdout, " -la local_address\n");
	fprintf(stdout, " -lp local_port\n");
	fprintf(stdout, " -br bucket_rate (in B/s)\n");
	fprintf(stdout, "\nDefault parameters are equivalent to:\n");
	fprintf(stdout, "%s -m b -sa ff02::11 -sp 3000 -la ::0 -lp 4000 -br 12000\n",
		prog_name);

}

///////////////////////////////////////////////////////////////
// report_error: print on stdout an introduction and an
// error code

void report_error( char *introduction, int err ) {
	if( (introduction != NULL) && (*introduction != '\0') ) { 
		fprintf( stdout, "%s: ", introduction );
	}
	if( err == 0 ) {
		fprintf( stdout, "No error\n" );
	} else {
		fprintf( stdout, "Error code: %d\n", err );
	}
}

///////////////////////////////////////////////////////////////
// printnl
int
printnl(void) {
	fprintf(stdout, "\n");
	return(0);
}
 
///////////////////////////////////////////////////////////////
// Create_Test_Socket
  
static 
SOCKET 
Create_Test_Socket( 
    void) 
 
//Routine Description: 
//     This  procedure  creates a socket to be used for the tests
//Arguments: 
//     none 
//Return Value: 
//    The  procedure  returns  the  newly  created  socket  if successful, 
//    otherwise it returns INVALID_SOCKET. 
{ 

    SOCKET  new_socket;

	WSAPROTOCOL_INFO rsvpipv6_info; // information on the rsvp over ipv6 prot.

	LPWSAPROTOCOL_INFO   proto_buf; 
    DWORD                size_guess; 
    BOOL                 still_trying; 
    BOOL                 got_protos; 
    int                  enum_result; 
    int                  error_code; 
	BOOL                 found_rsvpipv6;

	// initial values of the variables: 
    size_guess = 0; 
    got_protos = FALSE; 
    still_trying = TRUE; 
    proto_buf = NULL; 
 
    // We  call  WSCEnumProtocols  twice.   The first call determines how big a 
    // buffer  we will actually need for the results.  The second call actually 
    // retrieves  the  results.  In a situation such as this, I generally write 
    // the  sequence  of  two  calls  as  a  loop,  since this way the call and 
    // error-check code has to be written only once. 
    while (still_trying) { 
        enum_result = WSAEnumProtocols( 
            NULL,           // lpiProtocols (I want them all)
            proto_buf,      // lpProtocolBuffer 
            & size_guess);   // lpdwBufferLength 
		error_code = WSAGetLastError();
        if (enum_result != SOCKET_ERROR) { // I got them
            got_protos = TRUE; // I have them
            still_trying = FALSE; // so, don't ask again (out of while)
        } 
        else {
            if (error_code == WSAENOBUFS) { // buffer was too small
                delete [] proto_buf; 
                proto_buf = (LPWSAPROTOCOL_INFO) new char[size_guess]; 
                if (proto_buf == NULL) { // not enough memory. Bye!
                    still_trying = FALSE; // get out of while()
                } 
            } 
            else { // other error
                still_trying = FALSE; // get out of while
				report_error("Create_Test_Socket, WSAEnumProtocols", 
					error_code);
            } 
        } 
    } // while still_trying 

	// Look for rsvp over ipv6 (based on the provider id):

	found_rsvpipv6 = FALSE;
    if (got_protos) { // If I got the protocols:
        int  i; 
 
        for (i = 0; i < enum_result; i++) { 
            if (proto_buf[i].ProviderId == RSVPIPV6_ProviderId) {
                found_rsvpipv6 = TRUE; 
				rsvpipv6_info = proto_buf[i]; // remember information
                break; // out of for
            } 
        } 
    }

    delete [] proto_buf; // free memory used to learn about installed protocols
 
	
	if( found_rsvpipv6 == FALSE ) {
		WSASetLastError(WSAEPROVIDERFAILEDINIT);
		return( INVALID_SOCKET );
	}


	// rsvp over ipv6 found: create the socket

    new_socket = WSASocket( 
        FROM_PROTOCOL_INFO, // af 
        FROM_PROTOCOL_INFO, // type 
        FROM_PROTOCOL_INFO, // protocol 
        & rsvpipv6_info,    // lpProtocolInfo 
        0,                 // g (0 -> no group operation)
        0);                // dwFlags 
    return(new_socket); 
 
} // Create_Test_Socket 
 
 
///////////////////////////////////////////////////////////////
// Destroy_Test_Socket
  
static 
void 
Destroy_Test_Socket( 
    IN  SOCKET  TestSock 
    ) 
//Routine Description: 
// 
//    This  procedure  closes  a socket that was being used for tests
//
//Arguments: 
// 
//    TestSock - Supplies the socket to be closed. 
//     
//Return Value: 
// 
//    none 
{ 
    int    close_result; 
    DWORD  last_error; 

	fprintf(stdout, "test3, Destroy_Test_Socket called\n");
 
    close_result = closesocket(TestSock); 
    if (close_result != ERROR_SUCCESS) { 
        last_error = GetLastError(); 
        last_error = GetLastError();  // Redundant for single-step convenience. 
		report_error("test3, closesocket in Destroy_Test_Socket", 
			last_error);
    } 
     
} // Destroy_Test_Socket 
 

///////////////////////////////////////////////////////////////
// wait4key: auxiliar. For stopping until the user presses enter

void wait4key( void ) {

	fprintf(stdout, "Press enter to continue\n");
	fgetc(stdin);

}

///////////////////////////////////////////////////////////////
// poll4key: auxiliar. For checking if user pressed a key
// Note: it does not work!: select returns -1

BOOL poll4key( void ) {

	static fd_set keyboard;
	struct timeval zero_s = {0,0};
	int r;
	
	FD_ZERO(&keyboard);
	FD_SET(_fileno(stdin), &keyboard);

	r = select(_fileno(stdin),&keyboard, NULL, NULL,&zero_s);

//	fprintf(stdout, "select: %d\n", r);

	return( r > 0 );

}


///////////////////////////////////////////////////////////////
// functions for filling sample strucutres
///////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////
// fill_sample_local_ENDPOINT_ADDR
int fill_sample_local_ENDPOINT_ADDR( LPENDPOINT_ADDR lpea ) {

	char *straddr = source_addr;
	int res;
	
	lpea->sin_family = AF_INET6;
	lpea->sin_port = htons(source_port);

	res = inet6_addr(straddr, (struct in6_addr *)(&(lpea->sin_addr)));
	//fprintf(stdout, "inet6_addr returns %d\n", res);

	return( 0 );
}

///////////////////////////////////////////////////////////////
// fill_sample_session_ENDPOINT_ADDR
fill_sample_session_ENDPOINT_ADDR( LPENDPOINT_ADDR lpea ) {

	char *straddr = session_addr;
	ushort port = session_port;
	int res;
	
	lpea->sin_family = htons(AF_INET6);
	lpea->sin_port = htons(port);

	res = inet6_addr(straddr, (struct in6_addr *)(&(lpea->sin_addr)));
	//fprintf(stdout, "inet6_addr returns %d\n", res);

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_RSVP_SESSION

int fill_sample_RSVP_SESSION( LPRSVP_SESSION psess ) {

	fill_sample_session_ENDPOINT_ADDR( &(psess->addr) );
	psess->protId = IPPROTO_UDP;

	return( 0 );
}

///////////////////////////////////////////////////////////////
// fill_sample_RSVP_SENDER_TSPEC
int fill_sample_RSVP_SENDER_TSPEC( LPRSVP_SENDER_TSPEC lprst ) {
	
	lprst->type = enum_TokenBucket;
	lprst->length = RSVP_TSPEC_LENGTH(RSVP_TOKEN_BUCKET_TSPEC);
	lprst->tspec.tbTspec.tokenRate = tb_r ; // in bytes/sec
	lprst->tspec.tbTspec.bucketSize = tb_b; // in bytes
	lprst->tspec.tbTspec.peakRate = tb_p; // in bytes/sec
	lprst->tspec.tbTspec.minPoliced = tb_m; // in bytes
	lprst->tspec.tbTspec.maxPktSize = tb_M; // in bytes

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_RSVP_RECEIVER_TSPEC
int fill_sample_RSVP_RECEIVER_TSPEC( LPRSVP_RECEIVER_TSPEC lprrt ) {
	
	lprrt->type = enum_TokenBucket;
	lprrt->length = RSVP_TSPEC_LENGTH(RSVP_TOKEN_BUCKET_TSPEC);
	lprrt->tspec.tbTspec.tokenRate = 12000 ; // in bytes/sec
	lprrt->tspec.tbTspec.bucketSize = 4000; // in bytes
	lprrt->tspec.tbTspec.peakRate = 12000; // in bytes/sec
	lprrt->tspec.tbTspec.minPoliced = 512; // in bytes
	lprrt->tspec.tbTspec.maxPktSize = 512; // in bytes

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_RSVP_RECEIVER_RSPEC
int fill_sample_RSVP_RECEIVER_RSPEC( LPRSVP_RECEIVER_RSPEC lprrr ) {

	lprrr->type = enum_GuaranteedService; // the only possible type
	lprrr->length = RSVP_RSPEC_LENGTH(GS_RSPEC);
	lprrr->rspec.gsRspec.rate = 100;
	lprrr->rspec.gsRspec.slackTerm = 200;

	return(0);

}


///////////////////////////////////////////////////////////////
// fill_sample_SENDER_DATA
int fill_sample_SENDER_DATA( LPSENDER_DATA lpsd ) {

	fill_sample_RSVP_SESSION( &(lpsd->sess) );
	fill_sample_local_ENDPOINT_ADDR( &(lpsd->localAddr) );
	fill_sample_RSVP_SENDER_TSPEC( &(lpsd->tspec) );
	lpsd->lpPolicy = NULL;
	lpsd->dataTTL = 0;

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_REL_SENDER_DATA
int fill_sample_REL_SENDER_DATA( LPREL_SENDER_DATA lprsd ) {

	fill_sample_RSVP_SESSION( &(lprsd->sess) );
	fill_sample_local_ENDPOINT_ADDR( &(lprsd->sender) );

	return( 0 );

}



///////////////////////////////////////////////////////////////
// fill_sample_RSVP_FLOWSPEC_CL
int fill_sample_RSVP_FLOWSPEC_CL( LPRSVP_FLOWSPEC lprf ){

	lprf->type = enum_ControlledLoad;
	lprf->length = RSVP_FLOWSPEC_LENGTH(CL_FLOWSPEC);
	fill_sample_RSVP_RECEIVER_TSPEC( &(lprf->flowspec.clFlowspec.Tspec) );

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_RSVP_FLOWSPEC_GS
int fill_sample_RSVP_FLOWSPEC_GS( LPRSVP_FLOWSPEC lprf ){

	lprf->type = enum_Guaranteed;
	lprf->length = RSVP_FLOWSPEC_LENGTH(GS_FLOWSPEC);
	fill_sample_RSVP_RECEIVER_TSPEC( &(lprf->flowspec.gsFlowspec.Tspec) );
	fill_sample_RSVP_RECEIVER_RSPEC( &(lprf->flowspec.gsFlowspec.Rspec) );

	return( 0 );

}


///////////////////////////////////////////////////////////////
// fill_sample_RSVP_FILTERSPEC_0
int fill_sample_RSVP_FILTERSPEC_0( LPRSVP_FILTERSPEC lprf ){

	char straddr[STR_ADDR_LEN] = "fe80::1:a00:20ff:fe83:65f1";
	
	lprf->sin_family = htons(AF_INET6);
	lprf->sin_port = htons(1234);

	inet6_addr(straddr, (struct in6_addr *)(&(lprf->sin_addr)));
	
	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_RSVP_FILTERSPEC_1
int fill_sample_RSVP_FILTERSPEC_1( LPRSVP_FILTERSPEC lprf ){

	char straddr[STR_ADDR_LEN] = "fe80::1:a00:20ff:fe83:65f1";
	
	lprf->sin_family = AF_INET6;
	lprf->sin_port = htons(1234);

	inet6_addr(straddr, (struct in6_addr *)(&(lprf->sin_addr)));
		
	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_SE_RSVP_FLOWDESC
int fill_sample_SE_RSVP_FLOWDESC( LPRSVP_FLOWDESC lprf ) {

	fill_sample_RSVP_FLOWSPEC_CL( &(lprf->flow) );
	lprf->lpPolicy = NULL;
	lprf->numFilts = 2; // 2 filter specifications
	lprf->filtList = (LPRSVP_FILTERSPEC)
		malloc( (lprf->numFilts) * sizeof(RSVP_FILTERSPEC) );
	if(lprf->filtList == NULL) {
		fprintf(stdout, "test3: out of memory!\n");
		exit( 0 );
	}
	fill_sample_RSVP_FILTERSPEC_0( &((lprf->filtList)[0]) );
	fill_sample_RSVP_FILTERSPEC_1( &((lprf->filtList)[1]) );

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_WF_RSVP_FLOWDESC
int fill_sample_WF_RSVP_FLOWDESC( LPRSVP_FLOWDESC lprf ) {

	fill_sample_RSVP_FLOWSPEC_CL( &(lprf->flow) );
	//fill_sample_RSVP_FLOWSPEC_GS( &(lprf->flow) );
	lprf->lpPolicy = NULL;
	lprf->numFilts = 0; // no filter specifications (Wildcard style)
	lprf->filtList = NULL;

	return( 0 );

}


///////////////////////////////////////////////////////////////
// fill_sample_RESERVE_DATA_SE
int fill_sample_RESERVE_DATA_SE( LPRESERVE_DATA lprd ) {

	fill_sample_RSVP_SESSION( &(lprd->sess) );
	lprd->lpPolicy = NULL;
	lprd->style = RSVP_STYLE_SHARED_EXPLICIT;
	lprd->confmFlag = RESV_CONFIRM_REQUIRED; // I want confirmation
	fill_sample_local_ENDPOINT_ADDR( &(lprd->hostAddr) );
	lprd->numFds = 1;
	lprd->fdList = (LPRSVP_FLOWDESC)
		malloc((lprd->numFds)*sizeof(RSVP_FLOWDESC));
	if(lprd->fdList == NULL) {
		fprintf(stdout, "test3: out of memory!\n");
		exit( 0 );
	}
	fill_sample_SE_RSVP_FLOWDESC( lprd->fdList );

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_RESERVE_DATA_WF
int fill_sample_RESERVE_DATA_WF( LPRESERVE_DATA lprd ) {

	fill_sample_RSVP_SESSION( &(lprd->sess) );
	lprd->lpPolicy = NULL;
	lprd->style = RSVP_STYLE_WILDCARD_FILTER;
	lprd->confmFlag = RESV_CONFIRM_REQUIRED; // I want confirmation
	fill_sample_local_ENDPOINT_ADDR( &(lprd->hostAddr) );
	lprd->numFds = 1;
	lprd->fdList = (LPRSVP_FLOWDESC)
		malloc((lprd->numFds)*sizeof(RSVP_FLOWDESC));
	if(lprd->fdList == NULL) {
		fprintf(stdout, "test3: out of memory!\n");
		exit( 0 );
	}
	fill_sample_WF_RSVP_FLOWDESC( lprd->fdList );

	return( 0 );

}


///////////////////////////////////////////////////////////////
// fill_sample_REL_RESERVE_DATA_SE
int fill_sample_REL_RESERVE_DATA_SE( LPREL_RESERVE_DATA lprrd ) {

	fill_sample_RSVP_SESSION( &(lprrd->sess) );
	fill_sample_local_ENDPOINT_ADDR( &(lprrd->hostAddr) );
	lprrd->style = RSVP_STYLE_SHARED_EXPLICIT;
	lprrd->numFilts = 0; // filterspecs would be ignored anyway
	lprrd->filtList = NULL;

	return( 0 );

}

///////////////////////////////////////////////////////////////
// fill_sample_REL_RESERVE_DATA_WF
int fill_sample_REL_RESERVE_DATA_WF( LPREL_RESERVE_DATA lprrd ) {

	fill_sample_RSVP_SESSION( &(lprrd->sess) );
	fill_sample_local_ENDPOINT_ADDR( &(lprrd->hostAddr) );
	lprrd->style = RSVP_STYLE_WILDCARD_FILTER;
	lprrd->numFilts = 0; // filterspecs would be ignored anyway
	lprrd->filtList = NULL;

	return( 0 );

}


///////////////////////////////////////////////////////////////
// fill_sample_REGISTER_DATA
int fill_sample_REGISTER_DATA( LPREGISTER_DATA lprd ) {

	fill_sample_RSVP_SESSION( &(lprd->sess) );
	lprd->regOption = RSVP_SENRECV; // Sender and receiver

	return(0);

}

///////////////////////////////////////////////////////////////
// fill_sample_QUERY_QOS_DATA
int fill_sample_QUERY_QOS_DATA( LPQUERY_QOS_DATA lpqqd ) {

	fill_sample_RSVP_SESSION( &(lpqqd->sess) );
	// I don't fill up any more because the provider will
	// ignore it

	return(0);

}

int RSVP_RETURN_2_RESERVE_DATA(LPRSVP_RETURN retp, LPRESERVE_DATA resp) {

	ULONG i;

	resp->sess = retp->sess;
	resp->lpPolicy = NULL;
	resp->style = RSVP_STYLE_FIXED_FILTER;
	resp->confmFlag = RESV_CONFIRM_REQUIRED;
	fill_sample_local_ENDPOINT_ADDR(&(resp->hostAddr));
	
	resp->numFds = retp->infoLength;
	resp->fdList = new_flow_desc_list(resp->numFds);

	for(i=0; i < resp->numFds; i++) {

		(resp->fdList)[i].flow.type = enum_ControlledLoad;
		(resp->fdList)[i].flow.length = RSVP_FLOWSPEC_LENGTH(CL_FLOWSPEC);
		(resp->fdList)[i].flow.flowspec.clFlowspec.Tspec =
			(retp->ret.sndDesc)[i].tspec;

		(resp->fdList)[i].lpPolicy = NULL;

		(resp->fdList)[i].numFilts = 1; // fixed filter
		(resp->fdList)[i].filtList = new_RSVP_FILTERSPEC_list(1);
		(resp->fdList)[i].filtList[0] = (retp->ret.sndDesc)[i].sender;

	}

	return(0);
}

///////////////////////////////////////////////////////////////
// functions for printing results
///////////////////////////////////////////////////////////////

int print_ENDPOINT_ADDR( ENDPOINT_ADDR addr ) {

	char *addrstr;

	addrstr = inet6_ntoa(((struct in6_addr *)(&addr.sin_addr)));

	fprintf(stdout, "%s/%d", addrstr, ntohs(addr.sin_port));

	return(0);
}

int print_RSVP_SESSION( RSVP_SESSION sess ) {

	print_ENDPOINT_ADDR( sess.addr );
	fprintf(stdout, "[%d]", (int)sess.protId );

	return(0);
}

int print_RSVP_TSPEC( RSVP_TSPEC tspec ) {

	fprintf(stdout, "[r:%fB/s b:%fB p:%fB/s m:%dB M:%dB]",
		tspec.tspec.tbTspec.tokenRate,
		tspec.tspec.tbTspec.bucketSize,
		tspec.tspec.tbTspec.peakRate,
		tspec.tspec.tbTspec.minPoliced,
		tspec.tspec.tbTspec.maxPktSize);
	

	return(0);

}

int print_RSVP_RSPEC( RSVP_RSPEC rr ) {

	fprintf(stdout, "[R:%fB/s S:%dus]",
		rr.rspec.gsRspec.rate, 
		rr.rspec.gsRspec.slackTerm);

	return(0);

}

int print_RSVP_FLOWSPEC( RSVP_FLOWSPEC fs ) {

	fprintf(stdout, " type: ");
	switch(fs.type) {
	case(enum_Guaranteed):
		fprintf(stdout, "Guaranteed\n");
		break;
	case(enum_ControlledLoad):
		fprintf(stdout, "Controlled Load\n");
		break;
	default:
		fprintf(stdout, "Unknown\n");
		break;
	} // switch
	fprintf(stdout, " length: %d\n", fs.length);
	fprintf(stdout, " - Tspec:\n   ");
	switch(fs.type) {
	case(enum_ControlledLoad):
		print_RSVP_TSPEC(fs.flowspec.clFlowspec.Tspec);
		printnl();
		break;
	case(enum_Guaranteed):
		print_RSVP_TSPEC(fs.flowspec.gsFlowspec.Tspec);
		printnl();
		fprintf(stdout, " - Rspec:\n   ");
		print_RSVP_RSPEC(fs.flowspec.gsFlowspec.Rspec);
		printnl();
		break;
	} // switch


	return(0);

}

int print_RSVP_FLOWDESC( RSVP_FLOWDESC fd ) {

	ULONG i;
	
	fprintf(stdout, "Flow specification:\n");
	print_RSVP_FLOWSPEC( fd.flow );
	fprintf(stdout, "Number of filter specifications: %d\n", fd.numFilts);
	for( i=0; i < fd.numFilts; i++ ) {
		fprintf(stdout, "- Filter #%d:\n  ", i);
		print_ENDPOINT_ADDR(fd.filtList[i]);
		printnl();
	}

	return(0);

}

int print_RSVP_RETURN( RSVP_RETURN lprr ) {

	ULONG i;

	fprintf(stdout, "RSVP_RETURN structure obtained:\n");
	fprintf(stdout, "-- -- -- -- -- -- -- -- -- -- -\n");
	fprintf(stdout, "sess:\n  ");
	print_RSVP_SESSION( lprr.sess );
	printnl();
	fprintf(stdout, "infoLength: %d\n", lprr.infoLength);
	fprintf(stdout, "infoType: ");
	
	switch(lprr.infoType) {

	case(INFO_PATH):
		fprintf(stdout, "INFO_PATH\n");
		//lprr.infoLength = 1; // XXX for working with old emulaiton!!
		for(i=0; i < (ULONG)lprr.infoLength; i++) {
			fprintf(stdout, "sender description #%d:\n", i);
			fprintf(stdout, " - sender template: ");
			print_ENDPOINT_ADDR(lprr.ret.sndDesc[i].sender);
			printnl();
			fprintf(stdout, " - sender tspec: ");
			print_RSVP_TSPEC(lprr.ret.sndDesc[i].tspec);
			printnl();
		} // for
		break;

	case(INFO_PATH_TEAR):
		fprintf(stdout, "INFO_PATH_TEAR\n");
		// shouldn't have any sender descriptor
		break;

	case(INFO_PATH_ERR):
		fprintf(stdout, "INFO_PATH_ERR\n");
		fprintf(stdout, "eCode/eValue: %d/%d\n",
			(int)lprr.ret.pathErr->eCode, (int)lprr.ret.pathErr->eValue);
		fprintf(stdout, "Error node: ");
		print_ENDPOINT_ADDR(lprr.ret.pathErr->eNodeAddr); printnl();
		fprintf(stdout, "Sender Template: ");
		print_ENDPOINT_ADDR(lprr.ret.pathErr->eSender); printnl();
		break;

	case(INFO_RESV_ERR):
		fprintf(stdout, "INFO_RESV_ERR\n");
		fprintf(stdout, "eCode/eValue: %d/%d\n",
			(int)lprr.ret.resvErr->eCode, (int)lprr.ret.resvErr->eValue);
		fprintf(stdout, "Error node: ");
		print_ENDPOINT_ADDR(lprr.ret.resvErr->eNodeAddr); printnl();
		fprintf(stdout, "Number of flow descriptors: %d\n",
			lprr.ret.resvErr->numFds);
		for( i=0; i < lprr.ret.resvErr->numFds; i++ ) {
			fprintf(stdout, "** flow desc %d **\n", i);
			print_RSVP_FLOWDESC((lprr.ret.resvErr->fdList)[i]);
		}
		break;

	case(INFO_RESV_TEAR):
		fprintf(stdout, "INFO_RESV_TEAR\n");
		// filterspecifications:
		if(lprr.infoLength > 0) {

			fprintf(stdout, "Number of Filter Specifications: %d\n",
				lprr.ret.resvChg->numFilts);
			
			for(i=0; i < lprr.ret.resvChg->numFilts; i++) {
				fprintf(stdout, "- Filter #%d: ", i);
				print_ENDPOINT_ADDR(lprr.ret.resvChg->filtList[i]);
				printnl();					
			} // for
		
		} // if(lprr->infoLength > 0)
		break;

	case(INFO_RESV_CHANGE):
		fprintf(stdout, "INFO_RESV_CHANGE\n");
		print_RSVP_FLOWDESC(lprr.ret.resvChg[0]);
		break;

	case(INFO_QUERY_QOS):
		fprintf(stdout, "INFO_QUERY_QOS\n");
		fprintf(stdout, "session:\n  ");
		print_RSVP_SESSION(lprr.ret.queryQoS->sess);
		printnl();
		fprintf(stdout, "style: ");
		switch(lprr.ret.queryQoS->style) {
		case(RSVP_STYLE_FIXED_FILTER):
			fprintf(stdout, "RSVP_STYLE_FIXED_FILTER\n");
			break;
		case(RSVP_STYLE_SHARED_EXPLICIT):
			fprintf(stdout, "RSVP_STYLE_SHARED_EXPLICIT\n");
			break;
		case(RSVP_STYLE_WILDCARD_FILTER):
			fprintf(stdout, "RSVP_STYLE_WILDCARD_FILTER\n");
			break;
		default:
			fprintf(stdout, "Unknown\n");
			break;
		} // switch
		fprintf(stdout, "confmFlag: ");
		if(lprr.ret.queryQoS->confmFlag == RESV_CONFIRM_NOT_REQUIRED) {
			fprintf(stdout, "RESV_CONFIRM_NOT_REQUIRED\n");
		} else {
			fprintf(stdout, "RESV_CONFIRM_REQUIRED\n");
		}
		fprintf(stdout, "Host address:\n  ");
		print_ENDPOINT_ADDR(lprr.ret.queryQoS->hostAddr);
		printnl();
		fprintf(stdout, "Number of flow descriptors: %d\n",
			lprr.ret.queryQoS->numFds);
		for( i=0; i < lprr.ret.queryQoS->numFds; i++ ) {
			fprintf(stdout, "** flow desc %d **\n", i);
			print_RSVP_FLOWDESC((lprr.ret.queryQoS->fdList)[i]);
		}

		break;

	case(INFO_RESV_CONFIRM):
		fprintf(stdout, "INFO_RESV_CONFIRM\n");
		fprintf(stdout, "Source of conf.: ");
		print_ENDPOINT_ADDR(lprr.ret.resvConf->source); printnl();
		fprintf(stdout, "style: ");
		switch(lprr.ret.resvConf->style) {
		case(RSVP_STYLE_FIXED_FILTER):
			fprintf(stdout, "RSVP_STYLE_FIXED_FILTER\n");
			break;
		case(RSVP_STYLE_SHARED_EXPLICIT):
			fprintf(stdout, "RSVP_STYLE_SHARED_EXPLICIT\n");
			break;
		case(RSVP_STYLE_WILDCARD_FILTER):
			fprintf(stdout, "RSVP_STYLE_WILDCARD_FILTER\n");
			break;
		default:
			fprintf(stdout, "Unknown\n");
			break;
		} // switch

		fprintf(stdout, "Number of flow descriptors: %d\n",
			lprr.ret.resvConf->numFds);
		for( i=0; i < lprr.ret.resvConf->numFds; i++ ) {
			fprintf(stdout, "** flow desc %d **\n", i);
			print_RSVP_FLOWDESC((lprr.ret.resvConf->fdList)[i]);
		}

		break;

	default:
		fprintf(stdout, "Not recognized type\n");
		break;

	} // switch

	fprintf(stdout, "-------------------------------\n");
	return(0);

}

/////////////////////////////////////////////////////////////
// main
/////////////////////////////////////////////////////////////

main(int argc, char **argv) {
	
    SOCKET   test_sock1, test_sock2;//, test_sock3;
    int      start_return; 
    WSADATA  winsock_data; 
    DWORD    last_error = 0; 
    int      clean_return;
	int i; // auxiliar, for looping

	enum{send, rcvr, both} mode; // sender=0; receiver=1, both=2

//	char session_addr_def[STR_ADDR_LEN] = "ff02::11";
//	char source_addr_def[STR_ADDR_LEN] = "::0";
//	ushort session_port_def = 3000;
//	ushort source_port_def = 4000;

	WSAEVENT Event = NULL; // Event to be signaled when FD_QOS "happens"
	REGISTER_DATA RegisterData; // Register data to RSVP
	SENDER_DATA SenderData; // To be a sender
	REL_SENDER_DATA RelSenderData; // To cease being a sender
	RESERVE_DATA ReserveData; // To be a Receiver
	RESERVE_DATA ReserveData2; // to reserve against received path
	REL_RESERVE_DATA RelReserveData; // to tear down the sample reservation
	QUERY_QOS_DATA QueryQosData; // to get qos information
	RSVP_RETURN RsvpReturn; // Return structure for RSVP
	RSVP_ERROR RsvpError; // Error structure for RSVP calls

	DWORD BytesReturned; // No. of bytes returned (RSVP calls)
	int Resulti; // Store result of some int functions
	DWORD Resultdw; // Store result of some dword functions


	// Default values for program arguments:
	mode = both;
	session_addr = "ff02::11";
	source_addr = "::0";
	session_port = 3000;
	source_port = 4000;
	  // token bucket parameters:
	tb_r = 12000; // bytes/s
	tb_b = 4000; // bytes
	tb_p = 12000; // bytes/s
	tb_m = 512; // bytes
	tb_M = 1024; // bytes


	// Get arguments

	for(i = 1; i < argc; i++) {

		// mode
		// ----
		if( strcmp(argv[i], "-m") == 0 ) {
			// -m: next argument is the mode
			i++;
			if( i >= argc ) {
				Usage(argv[0]);
				return(1);
			}
			if( strcmp(argv[i], "s") == 0 ) {
				mode = send;
				fprintf(stdout, "Mode: sender\n");
				continue; // back to the beginning of the loop
			}
			if( strcmp(argv[i], "r") == 0 ) {
				mode = rcvr;
				fprintf(stdout, "Mode: receiver\n");
				continue; // back to the beginning of the loop
			}
			if( strcmp(argv[i], "b") == 0 ) {
				mode = both;
				fprintf(stdout, "Mode: both (sender & receiver)\n");
				continue; // back to the beginning of the loop
			}
			// Here? bad argument
			fprintf(stdout, "Bad argument: %s\n", argv[i]);
			Usage(argv[0]);
			return(1);
		}

		// session address
		// ---------------
		if( strcmp(argv[i], "-sa") == 0 ) {
			// -sa: next argument is the session address
			i++;
			if( i >= argc ) {
				Usage(argv[0]);
				return(1);
			}
			session_addr = argv[i];
			fprintf(stdout, "Session address: %s\n", session_addr);
			continue; // back to the beginning of the loop
		}

		// session port
		// ------------
		if( strcmp(argv[i], "-sp") == 0 ) {
			// -sp: next argument is the session port
			i++;
			if( i >= argc ) {
				Usage(argv[0]);
				return(1);
			}
			session_port = atoi(argv[i]);
			fprintf(stdout, "Session port: %d\n", session_port);
			continue; // back to the beginning of the loop
		}

		// local address
		// ---------------
		if( strcmp(argv[i], "-la") == 0 ) {
			// -la: next argument is the loocal address
			i++;
			if( i >= argc ) {
				Usage(argv[0]);
				return(1);
			}
			source_addr = argv[i];
			fprintf(stdout, "Local address: %s\n", source_addr);
			continue; // back to the beginning of the loop
		}

		// local port
		// ------------
		if( strcmp(argv[i], "-lp") == 0 ) {
			// -lp: next argument is the local port
			i++;
			if( i >= argc ) {
				Usage(argv[0]);
				return(1);
			}
			source_port = atoi(argv[i]);
			fprintf(stdout, "Local port: %d\n", source_port);
			continue; // back to the beginning of the loop
		}

		// bucket rate
		// ------------
		if( strcmp(argv[i], "-br") == 0 ) {
			// -r: next argument is the bucket rate
			i++;
			if( i >= argc ) {
				Usage(argv[0]);
				return(1);
			}
			tb_r = tb_b = tb_p = atof(argv[i]);
			continue; // back to the beginning of the loop
		}

		// help
		// ----
		if( strcmp(argv[i], "-h") == 0 ) {
			// -h:help wanted
			Usage(argv[0]);
			return(0);
		}

		// Here? bad argument
		fprintf(stdout, "Bad argument: %s\n", argv[i]);
		Usage(argv[0]);
		return(1);

	} // for(i = 1; i < argc; i++)

	// print the token bucket spec:
	fprintf(stdout, "Tspec = [r:%fB/s b:%fB p:%fB/s m:%dB M:%dB]\n",
		tb_r, tb_b, tb_p, tb_m, tb_M);


	// End of getting parameters
	// ---


	// Start up winsock2

	start_return = WSAStartup( 
        MAKEWORD(2,2),    // wVersionRequested 
        & winsock_data);  // lpWSAData 
 
	if(start_return != ERROR_SUCCESS) {
		last_error = WSAGetLastError(); 
        last_error = WSAGetLastError();  // Redundant for single-step convenience.         
		report_error( "test3, WSAStartup", last_error );
		return(0);
	}

	// here -> wsastartup successful

	// Create test socket(s)
	if( (test_sock1 = Create_Test_Socket()) == INVALID_SOCKET ) {
		last_error = WSAGetLastError();
		report_error( "test3, Create_Test_Socket", last_error );
		return(0);
	}

	fprintf( stdout, "test3: test_sock1 = %d\n",
		test_sock1);


	if( (test_sock2 = Create_Test_Socket()) == INVALID_SOCKET ) {
		last_error = WSAGetLastError();
		report_error( "test3, Create_Test_Socket", last_error );
		return(0);
	}


	// here -> socket creation successful


	//fprintf( stdout, "test3: test_sock2 = %d\n", test_sock2);


	// Create event
	if( (Event = WSACreateEvent()) == WSA_INVALID_EVENT ) {
		report_error("test3, WSACreateEvent", WSAGetLastError());
		Destroy_Test_Socket(test_sock1);
		WSACleanup();
		return(0);
	}

	// Fill in sample data for RSVP
	fill_sample_REGISTER_DATA( &RegisterData );
	fill_sample_SENDER_DATA( &SenderData );
	fill_sample_REL_SENDER_DATA( &RelSenderData );
	//fill_sample_RESERVE_DATA_SE( &ReserveData );
	fill_sample_RESERVE_DATA_WF( &ReserveData );
	//fill_sample_REL_RESERVE_DATA_SE( &RelReserveData );
	fill_sample_REL_RESERVE_DATA_WF( &RelReserveData );
	fill_sample_QUERY_QOS_DATA( &QueryQosData );


	// Register to RSVP
	Resulti = WSAIoctl(test_sock1,
		SIO_RSVP_REGISTER,
		(LPVOID)&RegisterData,
		sizeof(RegisterData),
		(LPVOID)&RsvpError,
		sizeof(RsvpError),
		&BytesReturned,
		NULL, NULL);
	if( Resulti == SOCKET_ERROR ) {
		report_error("test3, WSAIoctl for register", WSAGetLastError());
		Destroy_Test_Socket(test_sock1);
		WSACleanup();
		return(0);
	}


	// Register event notifications
	Resulti = WSAEventSelect(test_sock1,
		Event,
		(long)FD_QOS);
	if( Resulti == SOCKET_ERROR ) {
		report_error("test3, WSAEventSelect", WSAGetLastError());
		Destroy_Test_Socket(test_sock1);
		WSACleanup();
		return(0);
	}

	//Sleep(5000);
	wait4key();

	// Provide sender data
	if( mode != rcvr ) {
		Resulti = WSAIoctl(test_sock1,
			SIO_RSVP_SENDER,
			(LPVOID)&SenderData,
			sizeof(SenderData),
			(LPVOID)&RsvpError,
			sizeof(RsvpError),
			&BytesReturned,
			NULL, NULL);
		if( Resulti == SOCKET_ERROR ) {
			report_error("test3, WSAIoctl for sender", WSAGetLastError());
			Destroy_Test_Socket(test_sock1);
			WSACleanup();
			return(0);
		}
		
		//Sleep(5000);
		//wait4key();
	}

/*
	// Provide reservation data
	if( mode != send ) {
		Resulti = WSAIoctl(test_sock1,
			SIO_RSVP_RESERVE,
			(LPVOID)&ReserveData,
			sizeof(ReserveData),
			(LPVOID)&RsvpError,
			sizeof(RsvpError),
			&BytesReturned,
			NULL, NULL);
		if( Resulti == SOCKET_ERROR ) {
			report_error("test3, WSAIoctl for reserve", WSAGetLastError());
			Destroy_Test_Socket(test_sock1);
			WSACleanup();
			return(0);
		}
		
		//Sleep(5000);
		//wait4key();
	}
*/


	for( ; ; ) {
//	for( i=0; i < NUM_WAIT; i++ ) {
				
		// wait for notifications
		Resultdw = WSAWaitForMultipleEvents(1, // 1 event
			&Event, // pointer to the event
			FALSE, // return when ANY of the events is signaled
			(DWORD)TIMEOUT, // timeout in milliseconds
			FALSE);
		if( Resultdw == WSA_WAIT_FAILED ) {
			report_error("test3, WSAWaitForMultipleEvents", WSAGetLastError());
			Destroy_Test_Socket(test_sock1);
			WSACleanup();
			return(0);
		}
		if( Resultdw == WSA_WAIT_TIMEOUT ) {
			//report_error("test3, WSAWaitForMultipleEvents timed out",
			//	WSAGetLastError());
			continue;
		} else {
			report_error("test3, WSAWaitForMultipleEvents signals event",
				WSAGetLastError());
		}
				
				
		// reenable notifications
		do {
			// fprintf(stdout, "test3, issuing retrieve\n");
			Resulti = WSAIoctl(test_sock1,
				SIO_RSVP_RETRIEVE,
				NULL,
				0,
				(LPVOID)&RsvpReturn,
				sizeof(RsvpReturn),
				&BytesReturned,
				NULL, NULL);
			fprintf(stdout, "test3, retrieve made\n");
			if(Resulti != SOCKET_ERROR) {
				print_RSVP_RETURN(RsvpReturn);
				
				if(mode != send) {
					if( (RsvpReturn.infoType == INFO_PATH) &&
						(RsvpReturn.infoLength != 0) ) {
						// If path received, reserve against it:
						RSVP_RETURN_2_RESERVE_DATA(&RsvpReturn, &ReserveData2);
						Resulti = WSAIoctl(test_sock1,
							SIO_RSVP_RESERVE,
							(LPVOID)&ReserveData2,
							sizeof(ReserveData2),
							(LPVOID)&RsvpError,
							sizeof(RsvpError),
							&BytesReturned,
							NULL, NULL);
						if( Resulti == SOCKET_ERROR ) {
							report_error("test3, WSAIoctl for reserve", WSAGetLastError());
							Destroy_Test_Socket(test_sock1);
							WSACleanup();
							return(0);
						}
						// Free allocated memory:
						free_flow_desc_list(ReserveData2.numFds, 
							ReserveData2.fdList);
					}
				}

			}
			if( Resulti == SOCKET_ERROR ) {
				if(WSAGetLastError() == WSAEWOULDBLOCK){
					report_error("test3, Retrieved last pending event",
						WSAGetLastError());
				} else {
					report_error("test3, WSAIoctl for retrieve", WSAGetLastError());
					Destroy_Test_Socket(test_sock1);
					WSACleanup();
					return(0);
				}

			}
		} while( Resulti != SOCKET_ERROR );
				
		// Reset Event (necessary)
		WSAResetEvent( Event );

		
	} // for


	
	// Get qos information:
	Resulti = WSAIoctl(test_sock1,
		SIO_RSVP_QUERY_QOS,
		&(QueryQosData),
		sizeof(QueryQosData),
		(LPVOID)&RsvpReturn,
		sizeof(RsvpReturn),
		&BytesReturned,
		NULL, NULL);
	if( Resulti == SOCKET_ERROR ) {
		report_error("test3, WSAIoctl for Query QoS data", WSAGetLastError());
	} else {
		fprintf(stdout, "test3, query QoS made\n");
		print_RSVP_RETURN(RsvpReturn);
	}


	// Cease being a sender:
	if( mode != rcvr ) {
		Resulti = WSAIoctl(test_sock1, SIO_RSVP_REL_SENDER,
			(LPVOID)&RelSenderData, sizeof(RelSenderData),
			(LPVOID)&RsvpError, sizeof(RsvpError), &BytesReturned, NULL, NULL);
		if( Resulti == SOCKET_ERROR ) {
			report_error("test3, WSAIoctl for release sender", WSAGetLastError());
			Destroy_Test_Socket(test_sock1);
			WSACleanup();
			return(0);
		}
	}


	// Teardown the sample reservation:
	if( mode != send ) {
		Resulti = WSAIoctl(test_sock1, SIO_RSVP_REL_RESERVE,
			(LPVOID)&RelReserveData, sizeof(RelReserveData),
			(LPVOID)&RsvpError, sizeof(RsvpError), &BytesReturned, NULL, NULL);
		if( Resulti == SOCKET_ERROR ) {
			report_error("test3, WSAIoctl for release reservation", WSAGetLastError());
			Destroy_Test_Socket(test_sock1);
			WSACleanup();
			return(0);
		}
	}
	

	// Unregister event:
	Resulti = WSAEventSelect(test_sock1,
		NULL,
		(long)0);
	if( Resulti == SOCKET_ERROR ) {
		report_error("test3, WSAEventSelect for unregistering", WSAGetLastError());
		Destroy_Test_Socket(test_sock1);
		WSACleanup();
		return(0);
	}

	fprintf(stdout, "test3, successful WSAEventSelect to unregister event\n");

	// Close event
	if( Event != NULL ) {
		WSACloseEvent( Event );
		fprintf(stdout, "test3, closed event\n");
	}

	// Close socket:
	Destroy_Test_Socket(test_sock1);

	fprintf(stdout, "test3, after Destroy_Test_Socket\n");
		
	clean_return = WSACleanup();

	fprintf(stdout, "test3, WSACleanup returned: %d\n", clean_return);

	wait4key();

	return(0);


} // main 
